#include <memory>
#include <algorithm>
#include <string>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <deque>
#include <set>
#include <unordered_map>
#include <unordered_set>

#include <cstdlib>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>

#include "malloc.c"

#define DO_FORK

namespace fs = boost::filesystem;
namespace opts = boost::program_options;

/**
 * Configuration.
 */
struct Config {
    uint16_t port_ = 12001;
};

/**
 * Message.
 */
class Message {
public:
    explicit Message(
        const char* from,
        const char* subj,
        const char* text)
        : from_(from)
        , subj_(subj)
        , text_(text) {
    }

    ~Message(void) {
        if (from_) {
            dlfree(
                const_cast<void*>(
                    reinterpret_cast<const void*>(from_)));
        }

        if (subj_) {
            dlfree(
                const_cast<void*>(
                    reinterpret_cast<const void*>(subj_)));
        }

        if (text_) {
            dlfree(
                const_cast<void*>(
                    reinterpret_cast<const void*>(text_)));
        }
    }

    const char* from(void) const {
        return from_;
    }

    const char* subject(void) const {
        return subj_;
    }

    const char* text(void) const {
        return text_;
    }

private:
    const char* from_;
    const char* subj_;
    const char* text_;
};

/**
 * Exception.
 */
class Exception : public std::exception {
public:
    /**
     * Constructor.
     *
     * @param msg Message.
     */
    explicit Exception(const std::string& msg) : msg_(msg) {
    }

    virtual const char* what(void) const noexcept {
        return msg_.c_str();
    }

private:
    const std::string msg_;
};

/**
 * Exit with an error message.
 *
 * @param desc Options.
 * @param msg Message.
 */
static void ExitError(opts::options_description& desc, const std::string msg) {
    std::cout << desc << "\nERROR: " << msg << ".\n";
    exit(1);
}

/**
 * Parse arguments.
 *
 * @param argc Argument count.
 * @param argv Argument vector.
 * @param[out] conf Configuration.
 */
static void ParseArgs(int argc, char** argv, Config& conf) {
    std::stringstream usage;
    usage << "Usage: " PROJECT_NAME " <command> [options]";
    opts::options_description visible_desc(usage.str());

    try {
        opts::options_description common_desc("# Common options");
        common_desc.add_options()
            ("config", opts::value<std::string>(), "Path to configuration file.")
            ("help", "Print this usage information.")
            ("port", opts::value<uint16_t>(), "Port.")
            ;

        visible_desc.add(common_desc);

        opts::variables_map vm;
        opts::store(
            opts::command_line_parser(argc, argv)
                .options(visible_desc)
                .run(),
            vm);

        if (vm.count("config")) {
            fs::path path(vm["config"].as<std::string>());
            if (!fs::exists(path))
                ExitError(visible_desc, "configuration file does not exist");

            try {
                opts::store(opts::parse_config_file<char>(path.c_str(), visible_desc), vm);
            } catch (const std::exception& e) {
                ExitError(visible_desc, "unable to parse configuration file");
            }
        }

        opts::notify(vm);

        if (vm.count("port")) {
            conf.port_ = vm["port"].as<uint16_t>();
        }

        if (vm.count("help")) {
            std::cout << visible_desc;
            exit(0);
        }

    } catch (const std::exception& e) {
        std::stringstream msg;
        msg << "unable to process options -- " << e.what();
        ExitError(visible_desc, msg.str());
    }
}

/**
 * Set a read timeout.
 *
 * @param sk Socket.
 * @return True if successful.
 */
static bool SetReadTimeout(const int sk) {
    timeval tv;
    tv.tv_sec = 60;
    tv.tv_usec = 0;
    if (setsockopt(sk, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)) < 0) {
      std::cerr << "unable to set read timeout";
      return false;
    }

    return true;
}

/**
 * Read n bytes.
 *
 * @param sk Socket.
 * @param buf Buffer.
 * @param n Number of bytes to read.
 * @return True if successful.
 */
/*
static bool ReadBytes(const int sk, char* buf, const size_t n) {
    char* ptr = buf;
    while (ptr < buf + n) {
        if (!SetReadTimeout(sk)) {
            return false;
        }

        auto ret = recv(sk, ptr, ptr - buf + n, 0);
        if (ret <= 0) {
	  std::cerr << "unable to receive on socket";
            return false;
        }

        ptr += ret;
    }

    return true;
}
*/

/**
 * Read until a character is found.
 *
 * @param sk Socket.
 * @param buf Buffer.
 * @param n Number of bytes to read.
 * @param delim Delimiter.
 * @return True if successful.
 */
static bool ReadUntil(
    const int sk,
    char* buf,
    const size_t n,
    const char delim) {
    char* ptr = buf;
    while (ptr < buf + n - 1) {
        if (!SetReadTimeout(sk)) {
            return false;
        }

        auto ret = recv(sk, ptr, 1, 0);
        if (ret <= 0) {
	  std::cerr << "unable to receive on socket";
	  return false;
        }

        if (*ptr == delim) {
            ++ptr;
            *ptr = '\0';
            break;
        }

        ++ptr;
    }

    std::cerr << "read " << ptr - buf << " bytes";

    return ptr < buf + n - 1;
}

/**
 * Write n bytes.
 *
 * @param sk Socket.
 * @param buf Buffer.
 * @param n Number of bytes to write.
 * @return True if successful.
 */
static bool WriteBytes(const int sk, const char* buf, const size_t n) {
    auto ptr = buf;
    while (ptr < buf + n) {
        auto ret = send(sk, ptr, n - (ptr - buf), 0);
        if (ret <= 0) {
            std::cerr << "unable to send on socket";
            return false;
        }

        ptr += ret;
    }

    return true;
}

/**
 * Handle a client.
 *
 * @param sk Socket.
 */
static void OnClient(const int sk) {
    std::list<std::shared_ptr<Message>> msgs;

#define WriteConstant(x) WriteBytes(sk, x, sizeof(x))

    auto on_view_msgs = [&](void) {
        char buf[1024];
        WriteConstant("No problem!  Here are your messages.\n");

        size_t idx = 0u;
        for (const auto& msg : msgs) {
            ++idx;
            auto len = snprintf(buf, sizeof(buf), "Message %d\n", idx);
            WriteBytes(sk, buf, len);
            WriteConstant("From: ");
            WriteBytes(sk, msg->from(), strlen(msg->from()));
            WriteConstant("Subject: ");
            WriteBytes(sk, msg->subject(), strlen(msg->subject()));
            WriteConstant("Text: ");
            WriteBytes(sk, msg->text(), strlen(msg->text()));
        }
    };

    auto on_post_msg = [&](void) {
        char* from = reinterpret_cast<char*>(dlmalloc(64));
        char* subj = reinterpret_cast<char*>(dlmalloc(64));
        char* text = reinterpret_cast<char*>(dlmalloc(256));
        if (!from || !subj || !text) {
            WriteConstant("Sorry, we're full, try again later.\n");
            std::cerr << "unable to allocate message buffers";

            if (from) {
                dlfree(reinterpret_cast<void*>(from));
            }

            if (subj) {
                dlfree(reinterpret_cast<void*>(subj));
            }

            if (text) {
                dlfree(reinterpret_cast<void*>(text));
            }

            exit(0);
        }

        WriteConstant("From: ");
        ReadUntil(sk, from, 256, '\n');
        WriteConstant("Subject: ");
        ReadUntil(sk, subj, 256, '\n');
        WriteConstant("Text: ");
        ReadUntil(sk, text, 256, '\n');

        WriteConstant("Saving message...");
        msgs.emplace_back(std::make_shared<Message>(from, subj, text));
        WriteConstant("done.\n");
    };

    auto on_rm_msg = [&](void) {
        char buf[16];

        WriteConstant("Remove message #: ");
        ReadUntil(sk, buf, sizeof(buf), '\n');

        auto it = msgs.begin();
        auto idx = strtoul(buf, nullptr, 10) - 1;
        for (auto i = 0u; i < idx && it != msgs.end(); ++i, ++it) {
        }

        if (it == msgs.end()) {
            WriteConstant("Invalid message number.\n");
        } else {
            msgs.erase(it);
        }
    };

    auto on_report_bug = [&](void) {
        WriteConstant(
"I'm sorry, all of our operators are busy at the moment, please try again later.\n"
        );
        exit(127);
    };

    auto on_exit = [&](void) {
        WriteConstant("Have an awesome day!\n");
        exit(0);
    };

    auto on_unknown_input = [&](void) {
        WriteConstant("I'm sorry, what was that?\n");
    };

    char buf[16];
    const char menu[] =
"\n"
"Welcome to the Buggy Message Server!\n"
"\n"
"Please examine your options closely, as our menu has changed.\n"
"1) View all messages\n"
"2) Post a message\n"
"3) Remove a message\n"
"4) Report a bug\n"
"5) Exit\n"
"> ";

    while (true) {
        if (!WriteBytes(sk, menu, sizeof(menu))) {
            exit(0);
        }

        if (!ReadUntil(sk, buf, sizeof(buf), '\n')) {
            std::cerr << "invalid input";
            exit(0);
        }

        switch (buf[0]) {
            case '1':
                on_view_msgs();
                break;

            case '2':
                on_post_msg();
                break;

            case '3':
                on_rm_msg();
                break;

            case '4':
                on_report_bug();
                break;

            case '5':
                on_exit();
                break;

            default:
                on_unknown_input();
                break;
        }
    }
}

/**
 * Run the service.
 *
 * @param conf Configuration.
 */
static void RunService(Config& conf) {
    auto sk = socket(AF_INET, SOCK_STREAM, 0);
    if (sk < 0) {
        std::cerr << "unable to create server socket";
        std::cerr << strerror(errno);
        return;
    }

    sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(conf.port_);
    addr.sin_addr.s_addr = INADDR_ANY;

    auto opt = 1;
    if (setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
        std::cerr << "unable to set REUSE_ADDR on server socket";
        std::cerr << strerror(errno);
        return;
    }

    if (bind(sk, reinterpret_cast<sockaddr*>(&addr), sizeof(addr)) < 0) {
        std::cerr << "unable to bind server socket";
        std::cerr << strerror(errno);
        return;
    }

    if (listen(sk, 16) < 0) {
        std::cerr << "unable to listen on server socket";
        std::cerr << strerror(errno);
        return;
    }

    while (true) {
        sockaddr_in client_addr;
        socklen_t addr_len = sizeof(client_addr);
        memset(&client_addr, 0, sizeof(client_addr));
        auto client_sk =
            accept(
                sk,
                reinterpret_cast<sockaddr*>(&client_addr),
                &addr_len);
        if (client_sk < 0) {
            std::cerr << "unable to accept connection";
            std::cerr << strerror(errno);
            return;
        }

#ifdef DO_FORK
        pid_t child;
        switch (child = fork()) {
            case -1:
                std::cerr << "unable to fork client handler";
                std::cerr << strerror(errno);
                return;

            case 0:
#endif // DO_FORK
                OnClient(client_sk);
#ifdef DO_FORK
                exit(0);

            default:
#endif // DO_FORK
                close(client_sk);
#ifdef DO_FORK
                break;
        }

        while (true) {
            int st;
            if (waitpid(-1, &st, WNOHANG) < 0) {
                break;
            }
        }
#endif // DO_FORK
    }
}

/**
 * Main.
 */
int main(int argc, char** argv) {
    Config conf;
    ParseArgs(argc, argv, conf);
    RunService(conf);
    return 0;
}
